/*
 *  linux/arch/arm/mach-clps711x/irq.c
 *
 *  Copyright (C) 2000 Deep Blue Solutions Ltd.
 *  Copyright (C) 2001-2003 Atmark Techno, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/init.h>
#include <linux/list.h>
#include <linux/kernel.h>
#include <linux/timer.h>

#include <asm/mach/irq.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <asm/hardware/clps7111.h>

#ifdef CONFIG_ARCH_ARMADILLO
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <asm/arch/armadillo.h>
#endif

static void mask_irq_int1(unsigned int irq)
{
	u32 intmr1;

	intmr1 = clps_readl(INTMR1);
	intmr1 &= ~(1 << irq);
	clps_writel(intmr1, INTMR1);
}

static void mask_ack_irq_int1(unsigned int irq)
{
	u32 intmr1;

	intmr1 = clps_readl(INTMR1);
	intmr1 &= ~(1 << irq);
	clps_writel(intmr1, INTMR1);

	switch (irq) {
	case IRQ_CSINT:  clps_writel(0, COEOI);  break;
	case IRQ_TC1OI:  clps_writel(0, TC1EOI); break;
	case IRQ_TC2OI:  clps_writel(0, TC2EOI); break;
	case IRQ_RTCMI:  clps_writel(0, RTCEOI); break;
	case IRQ_TINT:   clps_writel(0, TEOI);   break;
	case IRQ_UMSINT: clps_writel(0, UMSEOI); break;
	}
}

static void unmask_irq_int1(unsigned int irq)
{
	u32 intmr1;

	intmr1 = clps_readl(INTMR1);
	intmr1 |= 1 << irq;
	clps_writel(intmr1, INTMR1);
}

static void mask_irq_int2(unsigned int irq)
{
	u32 intmr2;

	intmr2 = clps_readl(INTMR2);
	intmr2 &= ~(1 << (irq - 16));
	clps_writel(intmr2, INTMR2);
}

static void mask_ack_irq_int2(unsigned int irq)
{
	u32 intmr2;

	intmr2 = clps_readl(INTMR2);
	intmr2 &= ~(1 << (irq - 16));
	clps_writel(intmr2, INTMR2);

	switch (irq) {
	case IRQ_KBDINT: clps_writel(0, KBDEOI); break;
	}
}

static void unmask_irq_int2(unsigned int irq)
{
	u32 intmr2;

	intmr2 = clps_readl(INTMR2);
	intmr2 |= 1 << (irq - 16);
	clps_writel(intmr2, INTMR2);
}

#ifdef CONFIG_ARCH_ARMADILLO

static void mask_irq_intisa1 (unsigned int irq)
{
  miscreg_intmr1 |= 1 << (irq - IRQ_ISA_BASE - 8);
  MISCREG_INTMR1 = miscreg_intmr1;
}

static void mask_ack_irq_intisa1 (unsigned int irq)
{
  u8 mask;
  
  mask = 1 << (irq - IRQ_ISA_BASE - 8);
  miscreg_intmr1 |= mask;
  MISCREG_INTMR1 = miscreg_intmr1;
  
  MISCREG_INTCR1 = mask;
}

static void unmask_irq_intisa1 (unsigned int irq)
{
  miscreg_intmr1 &= ~(1 << (irq - IRQ_ISA_BASE - 8));
  MISCREG_INTMR1 = miscreg_intmr1;
}

static void mask_irq_intisa2 (unsigned int irq)
{
  miscreg_intmr2 |= 1 << (irq - IRQ_ISA_BASE - 4);
  MISCREG_INTMR2 = miscreg_intmr2;
}

static void mask_ack_irq_intisa2 (unsigned int irq)
{
  u8 mask;
  
  mask = 1 << (irq - IRQ_ISA_BASE - 4);
  miscreg_intmr2 |= mask;
  MISCREG_INTMR2 = miscreg_intmr2;
  
  MISCREG_INTCR2 = mask;
}

static void unmask_irq_intisa2 (unsigned int irq)
{
  miscreg_intmr2 &= ~(1 << (irq - IRQ_ISA_BASE - 4));
  MISCREG_INTMR2 = miscreg_intmr2;
}

static void mask_irq_intisa3 (unsigned int irq)
{
  miscreg_intmr3 |= 1 << (irq - IRQ_ISA_BASE);
  MISCREG_INTMR3 = miscreg_intmr3;
}

static void mask_ack_irq_intisa3 (unsigned int irq)
{
  u8 mask;
  
  mask = 1 << (irq - IRQ_ISA_BASE);
  miscreg_intmr3 |= mask;
  MISCREG_INTMR3 = miscreg_intmr3;
  
  MISCREG_INTCR3 = mask;
}

static void unmask_irq_intisa3 (unsigned int irq)
{
  miscreg_intmr3 &= ~(1 << (irq - IRQ_ISA_BASE));
  MISCREG_INTMR3 = miscreg_intmr3;
}

static void no_action (int irq, void *dev_id, struct pt_regs *regs)
{
}

static struct irqaction irq_isa = {
  name:    "isa",
  handler: no_action,
};

#endif

#ifdef CONFIG_FIQ

static void mask_irq_fiq1(unsigned int irq)
{
	u32 intmr1;

	intmr1 = clps_readl(INTMR1);
	intmr1 &= ~(1 << (irq - FIQ_START));
	clps_writel(intmr1, INTMR1);
}

static void mask_ack_irq_fiq1(unsigned int irq)
{
	u32 intmr1;

	intmr1 = clps_readl(INTMR1);
	intmr1 &= ~(1 << (irq - FIQ_START));
	clps_writel(intmr1, INTMR1);

	switch (irq - FIQ_START) {
	case FIQ_BLINT: clps_writel(0, BLEOI); break;
	case FIQ_MCINT: clps_writel(0, MCEOI); break;
	}
}

static void unmask_irq_fiq1(unsigned int irq)
{
	u32 intmr1;

	intmr1 = clps_readl(INTMR1);
	intmr1 |= 1 << (irq - FIQ_START);
	clps_writel(intmr1, INTMR1);
}

#ifdef INTMR3

static void mask_irq_fiq3(unsigned int irq)
{
	u32 intmr3;

	intmr3 = clps_readl(INTMR3);
	intmr3 &= ~(1 << (irq - FIQ_START - 8));
	clps_writel(intmr3, INTMR3);
}

static void unmask_irq_fiq3(unsigned int irq)
{
	u32 intmr3;

	intmr3 = clps_readl(INTMR3);
	intmr3 |= 1 << (irq - FIQ_START - 8);
	clps_writel(intmr3, INTMR3);
}

#endif

#endif

void __init clps711x_init_irq(void)
{
	unsigned int i;

	for (i = 0; i < NR_IRQS; i++) {
	        if (INT1_IRQS & (1 << i)) {
		        irq_desc[i].valid	= 1;
			irq_desc[i].probe_ok	= 1;
			irq_desc[i].mask_ack	= (INT1_ACK_IRQS & (1 << i)) ?
						   mask_ack_irq_int1 :
						   mask_irq_int1;
			irq_desc[i].mask	= mask_irq_int1;
			irq_desc[i].unmask	= unmask_irq_int1;
		}
		if (INT2_IRQS & (1 << i)) {
		        irq_desc[i].valid	= 1;
			irq_desc[i].probe_ok	= 1;
			irq_desc[i].mask_ack	= (INT2_ACK_IRQS & (1 << i)) ?
						   mask_ack_irq_int2 :
						   mask_irq_int2;
			irq_desc[i].mask	= mask_irq_int2;
			irq_desc[i].unmask	= unmask_irq_int2;
		}			
	}
#ifdef CONFIG_ARCH_ARMADILLO
	for (i = IRQ_ISA_BASE; i < NR_IRQS_ISA; i++) {
	  if (INTISA3_IRQS & (1 << (i - IRQ_ISA_BASE))) {
	    irq_desc[i].valid	 = 1;
	    irq_desc[i].probe_ok = 1;
	    irq_desc[i].mask_ack = mask_ack_irq_intisa3;
	    irq_desc[i].mask	 = mask_irq_intisa3;
	    irq_desc[i].unmask	 = unmask_irq_intisa3;
	  }
	  if (INTISA2_IRQS & (1 << (i - IRQ_ISA_BASE))) {
	    irq_desc[i].valid	 = 1;
	    irq_desc[i].probe_ok = 1;
	    irq_desc[i].mask_ack = mask_ack_irq_intisa2;
	    irq_desc[i].mask	 = mask_irq_intisa2;
	    irq_desc[i].unmask	 = unmask_irq_intisa2;
	  }
	  if (INTISA1_IRQS & (1 << (i - IRQ_ISA_BASE))) {
	    irq_desc[i].valid	 = 1;
	    irq_desc[i].probe_ok = 1;
	    irq_desc[i].mask_ack = mask_ack_irq_intisa1;
	    irq_desc[i].mask	 = mask_irq_intisa1;
	    irq_desc[i].unmask	 = unmask_irq_intisa1;
	  }
	}
#endif
#ifdef CONFIG_FIQ
	for (i = FIQ_START; i < NR_IRQS_FIQ; i++) {
	  if (FIQ1_IRQS & (1 << (i - FIQ_START))) {
	    irq_desc[i].valid	 = 1;
	    irq_desc[i].mask_ack = (FIQ1_ACK_IRQS & (1 << (i - FIQ_START))) ?
	                           mask_ack_irq_fiq1:
	                           mask_irq_fiq1;
	    irq_desc[i].mask	 = mask_irq_fiq1;
	    irq_desc[i].unmask	 = unmask_irq_fiq1;
	  }
#ifdef INTMR3
	  if (FIQ3_IRQS & (1 << (i - FIQ_START))) {
	    irq_desc[i].valid	 = 1;
	    irq_desc[i].mask_ack = mask_irq_fiq3;
	    irq_desc[i].mask	 = mask_irq_fiq3;
	    irq_desc[i].unmask	 = unmask_irq_fiq3;
	  }
	}
#endif
#endif

	/*
	 * Disable interrupts
	 */
	clps_writel(0, INTMR1);
	clps_writel(0, INTMR2);
#ifdef CONFIG_ARCH_ARMADILLO
	miscreg_intmr1 = 0x7;
	MISCREG_INTMR1 = miscreg_intmr1;
	miscreg_intmr2 = 0xf;
	MISCREG_INTMR2 = miscreg_intmr2;
	miscreg_intmr3 = 0xf;
	MISCREG_INTMR3 = miscreg_intmr3;
#endif
#ifdef CONFIG_FIQ
#ifdef INTMR3
	clps_writel(0, INTMR3);
#endif
#endif

	/*
	 * Clear down any pending interrupts
	 */
	clps_writel(0, COEOI);
	clps_writel(0, TC1EOI);
	clps_writel(0, TC2EOI);
	clps_writel(0, RTCEOI);
	clps_writel(0, TEOI);
	clps_writel(0, UMSEOI);
	clps_writel(0, SYNCIO);
	clps_writel(0, KBDEOI);
#ifdef CONFIG_ARCH_ARMADILLO
	MISCREG_INTCR1 = 0x7;
	MISCREG_INTCR2 = 0xf;
	MISCREG_INTCR3 = 0xf;
	
	setup_arm_irq (IRQ_EINT1, &irq_isa);
#endif
#ifdef CONFIG_FIQ
	clps_writel(0, BLEOI);
	clps_writel(0, MCEOI);
	
	init_FIQ ();
#endif
}
