/*
 * linux/driver/ide/arm/ide-armadillo.c
 *
 * Copyright (C) 2001-2003 Atmark Techno, Inc.
 */

#include <linux/config.h>
#include <linux/ide.h>
#include <asm/arch/armadillo.h>

#include "../ide_modes.h"

static int miscreg_pio = -1;

static void armadillo_ide_tuneproc (ide_drive_t *drive, u8 pio)
{
  int pio_mode;

  if (drive->id) {
    if ((pio_mode = ide_scan_pio_blacklist(drive->id->model)) == -1) {
      if ((drive->id->field_valid & 2) && (drive->id->capability & 8) &&
	  (drive->id->eide_pio_modes & 7)) {
	    pio_mode = 3;
      } else {
	pio_mode = 0;
      }
    } else {
      pio_mode = (pio_mode < 3) ? 0 : 3;
    }
  } else {
    pio_mode = 0;
  }
  if (pio_mode >= 3) {
    if (ide_config_drive_speed (drive, XFER_PIO_3)) {
      printk ("ide-armadillo: ide_config_drive_speed error.\n");
      pio_mode = 0;
    }
  }
  if (pio_mode < 3) {
    if (miscreg_pio != 0) {
      miscreg_pio = 0;
      miscreg_idemode &= ~MISCREG_IDEPIO3;
      MISCREG_IDEMODE = miscreg_idemode;
      printk ("ide-armadillo: Set PIO mode 0.\n");
    }
  } else {
    if (miscreg_pio < 0) {
      miscreg_pio = 3;
      miscreg_idemode |= MISCREG_IDEPIO3;
      MISCREG_IDEMODE = miscreg_idemode;
      printk ("ide-armadillo: Set PIO mode 3.\n");
    }
  }
}

void armadillo_ide_init_hwifs (void)
{
  int i, j;
  
  for (i = 0; i < MAX_HWIFS; i++) {
    for (j = 0; j < MAX_DRIVES; j++) {
      ide_hwifs[i].drives[j].autotune = IDE_TUNE_AUTO;
    }
    ide_hwifs[i].tuneproc = &armadillo_ide_tuneproc;
  }
}
